"""
Constants and functions for use across the Becca core
"""

from __future__ import print_function
import os
import sys
import matplotlib.pyplot as plt
import numpy as np

# Shared constants
epsilon = sys.float_info.epsilon
big = 10 ** 20
max_int16 = np.iinfo(np.int16).max

# Colors for plotting
dark_grey = (0.2, 0.2, 0.2)
light_grey = (0.9, 0.9, 0.9)
red = (0.9, 0.3, 0.3)
# Becca pallette
copper_highlight = (253./255., 249./255., 240./255.)
light_copper = (242./255., 166./255., 108./255.)
copper = (175./255., 102./255, 53./255.)
dark_copper = (132./255., 73./255., 36./255.)
copper_shadow = (25./255., 22./255, 20./255.)
oxide = (20./255., 120./255., 150./255.)


def pad(arr, shape, val=0., dtype=float):
    """
    Pad a numpy array to the specified shape.

    Use val (default 0) to fill in the extra spaces.

    Parameters
    ----------
    arr : array of ints or floats
        The array to pad.
    shape : int, list of ints or tuple of ints
        The shape to which to pad ``a``.
        If any element of shape is 0, that size remains unchanged in
        that axis. If any element of shape is < 0, the size of ``a`` in that
        axis is incremented by the magnitude of that value.
    val : float
        The value with which to pad ``arr``. Default is 0.
    dtype : dtype
        The data type with which to pad ``arr``.

    Returns
    -------
    padded : array of ints or floats
        The padded version of ``arr``.
    """
    # For  padding a 1D array
    if isinstance(shape, (int, long)):
        if shape <= 0:
            rows = arr.size - shape
        else:
            rows = shape
            if rows < arr.size:
                print(' '.join(['arr.size is', str(arr.size),
                                ' but trying to pad to ',
                                str(rows), 'rows.']))
                return arr
        # Handle the case where a is a one-dimensional array
        padded = np.ones(rows, dtype=dtype) * val
        padded[:arr.size] = arr
        return padded

    # For padding arr n-D array
    new_shape = shape
    n_dim = len(shape)
    if n_dim > 4:
        print(''.join([str(n_dim), ' dimensions? Now you\'re getting greedy']))
        return arr

    for dim, _ in enumerate(shape):
        if shape[dim] <= 0:
            new_shape[dim] = arr.shape[dim] - shape[dim]
        else:
            if new_shape[dim] < arr.shape[dim]:
                print(''.join(['The variable shape in dimension ',
                               str(dim), ' is ', str(arr.shape[dim]),
                               ' but you are trying to pad to ',
                               str(new_shape[dim]), '.']))
                print('You aren\'t allowed to make it smaller.')
                return arr

    padded = np.ones(new_shape, dtype=dtype) * val
    if len(new_shape) == 2:
        padded[:arr.shape[0], :arr.shape[1]] = arr
        return padded
    if len(new_shape) == 3:
        padded[:arr.shape[0], :arr.shape[1], :arr.shape[2]] = arr
        return padded
    # A maximum of 4 dimensions is enforced.
    padded[:arr.shape[0], :arr.shape[1], :arr.shape[2], :arr.shape[3]] = arr
    return padded


def str_to_int(exp):
    """
    Convert a string to an integer.

    The method is primitive, using a simple hash based on the
    ordinal value of the characters and their position in the string.

    Parameters
    ----------
    exp : str
        The string expression to convert to an int.

    Returns
    -------
    sum : int
        An integer that is likely (though not extremely so) to be unique
        within the scope of the program.
    """
    sum_ = 0
    for i, character in enumerate(exp):
        sum_ += i + ord(character) + i * ord(character)
    return sum_


def timestr(timestep, s_per_step=.25, precise=True):
    """
    Convert the number of time steps into an age.

    Parameters
    ----------
    timestep : int
        The age in time steps.
    s_per_step : float
        The duration of each time step in seconds.
    precise : bool
        If True, report the age down to the second.
        If False, just report the most significant unit of time.
        Default is True

    Returns
    -------
    time_str : str
        The age in string format, including, as appropriate, years,
        months, days, hours, minutes, and seconds.
    """
    # Start by calculating the total number of seconds.
    total_sec = timestep * s_per_step
    sec = int(np.mod(total_sec, 60.))
    time_str = ' '.join([str(sec), 'sec'])

    # If necessary, calculate the total number of minutes.
    total_min = int(total_sec / 60)
    if total_min == 0:
        return time_str
    min_ = int(np.mod(total_min, 60.))
    if precise:
        time_str = ' '.join([str(min_), 'min', time_str])
    else:
        time_str = ' '.join([str(min_), 'min'])

    # If necessary, calculate the total number of hours.
    total_hr = int(total_min / 60)
    if total_hr == 0:
        return time_str
    hr_ = int(np.mod(total_hr, 24.))
    if precise:
        time_str = ' '.join([str(hr_), 'hr', time_str])
    else:
        time_str = ' '.join([str(hr_), 'hr'])

    # If necessary, calculate the total number of days.
    total_day = int(total_hr / 24)
    if total_day == 0:
        return time_str
    day = int(np.mod(total_day, 30.))
    if precise:
        time_str = ' '.join([str(day), 'dy', time_str])
    else:
        time_str = ' '.join([str(day), 'dy'])

    # If necessary, calculate the total number of months.
    total_mon = int(total_day / 30)
    if total_mon == 0:
        return time_str
    mon = int(np.mod(total_mon, 12.))
    if precise:
        time_str = ' '.join([str(mon), 'mo', time_str])
    else:
        time_str = ' '.join([str(mon), 'mo'])

    # If necessary, calculate the total number of years.
    yr_ = int(total_mon / 12)
    if yr_ == 0:
        return time_str
    if precise:
        time_str = ' '.join([str(yr_), 'yr', time_str])
    else:
        time_str = ' '.join([str(yr_), 'yr'])

    return time_str


def format_decimals(array):
    """
    Format and print an array as a list of fixed decimal numbers in a string.

    Parameters
    ----------
    array : array of floats
        The array to be formatted.
    """
    if len(array.shape) == 2:
        for j in range(array.shape[1]):
            formatted = (' '.join(['{0},{1}:{2:.3}'.format(i, j, array[i, j])
                                   for i in range(array.shape[0])]))
            print(formatted)
    else:
        array = array.copy().ravel()
        formatted = (' '.join(['{0}:{1:.3}'.format(i, array[i])
                               for i in range(array.size)]))
        print(formatted)


def get_files_with_suffix(dir_name, suffixes):
    """
    Get all of the files with a given suffix in dir recursively.

    Parameters
    ----------
    dir_name : str
        The path to the directory to search.
    suffixes : list of str
        The set of suffixes for which files are being collected.

    Returns
    -------
    found_filenames : list of str
        The filenames, including the local path from ``dir_name``.
    """
    found_filenames = []
    for localpath, _, filenames in os.walk(dir_name):
        for filename in filenames:
            for suffix in suffixes:
                if filename.endswith(suffix):
                    found_filenames.append(os.path.join(localpath, filename))
    found_filenames.sort()
    return found_filenames


def visualize_array(image_data, label='data_figure'):
    """
    Produce a visual representation of the image_data matrix.

    Parameters
    ----------
    image_data : 2D array of floats
        The pixel values to make into an image.
    label : str
        The string label to affix to the image. It is used both
        to generate a figure number and as the title.
    """
    # Treat nan values like zeros for display purposes
    image_data = np.nan_to_num(np.copy(image_data))

    fig = plt.figure(str_to_int(label))
    # Diane made the brilliant suggestion to leave this plot in color.
    # It looks much prettier.
    plt.summer()
    img = plt.imshow(image_data)
    img.set_interpolation('nearest')
    plt.title(label)
    plt.xlabel('Max = {0:.3}, Min = {1:.3}'.format(np.max(image_data),
                                                   np.min(image_data)))
    fig.show()
    fig.canvas.draw()



"""
The Affect class.
"""

import os
import numpy as np
import matplotlib.pyplot as plt

import becca.tools as tools


class Affect(object):
    """
    Assign reward to the appropriate features and track it over time.

    Affect or mood, is the level of arousal of the brain.
    It is influenced by the recent history of reward and in turn
    influences the intensity with which a brain pursues
    its goals and makes plans.
    """
    def __init__(self):
        """
        Set up Affect.

        Parameters
        ----------
        num_features : int
            The number of features in the brain. Affect will
            learn the reward associated with each of these.
        """
        # satisfaction_time_constant : float
        #     The time constant of the leaky integrator used to filter
        #     reward into a rough average of the recent reward history.
        self.satisfaction_time_constant = 1e3
        # satisfaction : float
        #     A filtered average of the reward.
        self.satisfaction = 0.
        # cumulative_reward : float
        #     The total reward amassed since the last visualization.
        self.cumulative_reward = 0.
        # time_since_reward_log : int
        #     Number of time steps since reward was last logged. It gets
        #     logged every time Affect is visualized.
        self.time_since_reward_log = 0.
        # reward_history : list of floats
        #     A time series of reward accumulated during the periods between
        #     each time Affect is visualized..
        self.reward_history = []
        # reward_steps : list of ints
        #     A time series of the brain's age in time steps corresponding
        #     to each of the rewards in reward_history.
        self.reward_steps = []


    def update(self, reward):
        """
        Update the current level of satisfaction and record the reward.

        Parameters
        ----------
        reward : float
            The most recently observed reward value.

        Returns
        -------
        self.satisfaction : float
        """
        # Clip the reward so that it falls between -1 and 1.
        reward = np.maximum(np.minimum(reward, 1.), -1.)

        # Update the satisfaction, a filtered version of the reward.
        rate = 1. / self.satisfaction_time_constant
        # This filter is also known as a leaky integrator.
        self.satisfaction = self.satisfaction * (1. - rate) + reward * rate

        # Log the reward.
        self.cumulative_reward += reward
        self.time_since_reward_log += 1

        return self.satisfaction


    def visualize(self, timestep, brain_name, log_dir):
        """
        Update the reward history, create plots, and save them to a file.

        Parameters
        ----------
        timestep : int
            See docstring for brain.py.
        brain_name : str
            See docstring for brain.py.
        log_dir : str
            See docstring for brain.py.

        Returns
        -------
        performance : float
            The average reward over the lifespan of the brain.
        """
        # Check whether any time has passed since the last update.
        if self.time_since_reward_log > 0:
            # Update the lifetime record of the reward.
            self.reward_history.append(float(self.cumulative_reward) /
                                       (self.time_since_reward_log + 1))
            self.cumulative_reward = 0
            self.time_since_reward_log = 0
            self.reward_steps.append(timestep)

        performance = np.mean(self.reward_history)

        # Plot the lifetime record of the reward.
        fig = plt.figure(11111)
        color = (np.array(tools.copper) +
                 np.random.normal(size=3, scale=.1))
        color = np.maximum(np.minimum(color, 1.), 0.)
        color = tuple(color)
        linewidth = np.random.normal(loc=2.5)
        linewidth = 2
        linewidth = np.maximum(1., linewidth)
        plt.plot(self.reward_steps, self.reward_history, color=color,
                 linewidth=linewidth)
        plt.gca().set_axis_bgcolor(tools.copper_highlight)
        plt.xlabel('Time step')
        plt.ylabel('Average reward')
        plt.title('Reward history for {0}'.format(brain_name))
        fig.show()
        fig.canvas.draw()

        # Save a copy of the plot.
        filename = 'reward_history_{0}.png'.format(brain_name)
        pathname = os.path.join(log_dir, filename)
        plt.savefig(pathname, format='png')

        return performance



"""
Node functions.

This module contains a couple of functions that do most of the heavy lifting.
They use numba to compile int C, so they run really fast. Even faster
than vectorized numpy. But not as fast as a GPU.

step() : This one is repsponsible for the one-time step advancement
    of a node. It is the heart of Becca. It recursively calls all the
    other nodes within a level too. Because recurive calls within numba
    is a relatively new feature, this demands numba 27 or later.
    It is also the reason that the function signature is explicitly
    typed, instead of just using an @autojit decorator.

update_rewards() : This function updates the reward associated with
    each of the nodes. It manages the credit assignment and reward updates
    across all nodes--some of the most critical aspects of Becca's
    reinforcement learning algorithm.
"""

import numba as nb
import numpy as np

@nb.jit(nb.types.Tuple((nb.int32, nb.int32))
        (nb.int32, #node_index
         nb.float64[:], #activity,
         nb.float64[:], #prev_activities,
         nb.float64, #activity_rate,
         nb.float64[:], #cumulative_activities,
         nb.float64[:], #attempts,
         nb.float64[:], #fulfillment,
         nb.float64[:], #unfulfillment,
         nb.float64, #curiosity_rate,
         nb.float64[:], #curiosity,
         nb.float64[:], #reward,
         nb.int32[:], #element_index,
         nb.int32[:], #sequence_index,
         nb.int32[:], #sequence_length,
         nb.float64, #bud_threshold,
         nb.int32[:], #num_children,
         nb.int32[:, :], #child_indices,
         nb.int32[:], #parent_index,
         nb.float64[:], #element_activities
         nb.int32, #num_active_elements,
         nb.float64, #parent_activity,
         nb.float64, #prev_parent_activity,
         nb.float64[:], #sequence_activities,
         nb.int32, #num_nodes,
         nb.int32, #num_sequences,
         nb.int32, #max_num_sequences,
         nb.int32, #max_num_nodes,
         nb.float64[:], #goal_votes,
         nb.int32[:], #responsible_nodes,
         nb.float64[:], #element_goals,
         nb.float64[:], #sequence_goals,
         nb.float64, #new_reward,
         nb.float64 #satisfaction
        ), nopython=True)

def step(node_index, # node parameters
         activity,
         prev_activities,
         activity_rate,
         cumulative_activities,
         attempts,
         fulfillment,
         unfulfillment,
         curiosity_rate,
         curiosity,
         reward,
         element_index,
         sequence_index,
         sequence_length,
         sequence_threshold,
         num_children,
         child_indices,
         parent_index,
         element_activities, # level parameters
         num_active_elements,
         parent_activity,
         prev_parent_activity,
         sequence_activities,
         num_nodes,
         num_sequences,
         max_num_sequences,
         max_num_nodes,
         goal_votes,
         responsible_nodes,
         element_goals,
         sequence_goals,
         new_reward,
         satisfaction
        ):
    """
    Update the Node with the most recent element activities.

    The step() function carries out Becca's feature creation,
    model learning and action selection. These are all incremental
    methods, meaning that they take the new inputs at each time step
    and update all the internal parameters and estimates incrementally.

    I apologize for the ridiculous number of variables passed into this
    function. When I figure out a more elegant way to do this, I'll
    implement it. Most of the input arguments have the same name as
    in level.py and are defined there.
    Node-related parameters, have an extra prefix, 'node_'.
    The few exceptions to that are defined here.

    Parameters
    ----------------
    new_reward : float
        The current value of the reward.
    parent_activity, prev_parent_activity : float
        The activity of the this node's parent from the current and previous
        time step..
    """
    # For the duration of this function call, the processing is focused on
    # node i.
    i = node_index

    # Note what the attempt was from the previous time step.
    # element_goals havent been updated yet, so they still
    # represent what happened on the previous time step.
    # prev_parent_activity (parent activity from the previous time step)
    # is used instead of the current parent
    # activity in order to introduce the one-time-step-per-element
    # temporal structure of the sequence.
    attempts[i] = min(prev_parent_activity, element_goals[element_index[i]])

    # Calculate the new node activity.
    # Node activity is either the incoming activity associated with
    # the element index, or the activity from the upstream node,
    # whichever is less.
    if i == 0:
        activity[i] = 1.
    else:
        activity[i] = min(prev_parent_activity,
                          element_activities[element_index[i]])
    cumulative_activities[i] += activity[i]

    # Fulfill curiosity on the previous time step's attempts.
    # Don't fulfill curiosity on activity. Activity may or may not occur,
    # but attempts the agent can control.
    curiosity[i] -= attempts[i]
    curiosity[i] = max(0., curiosity[i])

    # Increment the curiosity based on several multiplicative factors.
    #     curiosity_rate : a constant
    #     uncertainty : an estimate of how much is not yet known about
    #         this node. It is a function of the total number of past
    #         attempts, both fulfilled and unfulfilled.
    #     parent_activity : The activity of the parent node. If the
    #         parent isn't active, then the sequence represented by the
    #         child isn't active either.
    #     1 - curiosity : This is a squashing factor that ensures that
    #         curiosity will asymptotically approach 1.
    #     1 - satisfaction : This is a scaling factor to account for
    #         contentment in the agent. If the agent is consistently
    #         getting high rewards, there is little need to be curious.
    uncertainty = 1. / (1. + fulfillment[i] + unfulfillment[i])
    curiosity[i] += (curiosity_rate *
                     uncertainty *
                     parent_activity *
                     (1. - curiosity[i]) *
                     (1. - satisfaction))

    # Fulfillment is when an attempt is followed by node activity.
    new_fulfillment = min(activity[i], attempts[i])
    fulfillment[i] += new_fulfillment
    attempts[i] -= new_fulfillment

    # Unfulfuillment is when an attempt is not followed by node activity.
    # Decay any not-yet fulfilled goals.
    decay = attempts[i] * activity_rate
    unfulfillment[i] += decay
    attempts[i] -= decay

    # Get the goal value passed down from the level above.
    if sequence_index[i] != -1:
        top_down_goal = sequence_goals[sequence_index[i]]
    else:
        top_down_goal = 0.

    # If this node has a sequence associated with it
    # set the relevant sequence activity.
    if sequence_index[i] != -1:
        sequence_activities[sequence_index[i]] = activity[i]

    # If this is not the root node or a first generation child,
    # set the relevant element goal vote.
    if sequence_length[i] > 1:
        # The element goal vote is a combination of curiosity,
        # reward, and any temporary internal goal.
        # This node is only relevant when the parent is active.
        element_goal_vote = parent_activity * (curiosity[i] +
                                               reward[i] +
                                               top_down_goal)
        # For each element, keep the maximum vote it receives from
        # across all the nodes. Also note the node that cast the
        # highest vote. This helps explain what Becca is doing
        # for visualization and debugging.
        if element_goal_vote > goal_votes[element_index[i]]:
            goal_votes[element_index[i]] = element_goal_vote
            responsible_nodes[element_index[i]] = i

    # Do a depth-first recursive traversal of the tree. Explore each
    # of the child nodes, one at a time.
    if num_children[i] > 0:
        parent_activity = activity[i]
        prev_parent_activity = prev_activities[i]
        for child_index in child_indices[i, :num_children[i]]:
            # Only explore child nodes corresponding to active elements.
            if element_index[child_index] < num_active_elements:
                num_nodes, num_sequences = (
                    step(child_index, # node parameters
                         activity,
                         prev_activities,
                         activity_rate,
                         cumulative_activities,
                         attempts,
                         fulfillment,
                         unfulfillment,
                         curiosity_rate,
                         curiosity,
                         reward,
                         element_index,
                         sequence_index,
                         sequence_length,
                         sequence_threshold,
                         num_children,
                         child_indices,
                         parent_index,
                         element_activities, # level parameters
                         num_active_elements,
                         parent_activity,
                         prev_parent_activity,
                         sequence_activities,
                         num_nodes,
                         num_sequences,
                         max_num_sequences,
                         max_num_nodes,
                         goal_votes,
                         responsible_nodes,
                         element_goals,
                         sequence_goals,
                         new_reward,
                         satisfaction))

    # If there is still room for more sequences, grow them.
    else: #if num_children[i] == 0:
        if num_sequences < max_num_sequences:
            # As the number of sequences increases, boost the threshold.
            # This makes it increasingly harder to create new sequences.
            # The early ones are created quickly and later ones, more slowly.
            # This balances an aggresive rapid-learning strategy with a
            # more conservative slow-learning approach.
            mod_threshold = sequence_threshold * (
                1. + num_sequences / max_num_sequences)
            if cumulative_activities[i] > mod_threshold:
                # Assign this node its own output sequence, now that it
                # has been observed enough times.
                sequence_index[i] = num_sequences
                num_sequences += 1

    # Pass the node activity to the next time step.
    prev_activities[i] = activity[i]

    return num_nodes, num_sequences


@nb.jit(nb.types.Tuple((nb.float64[:], nb.float64[:, :], nb.int32))
        (nb.float64[:], #node_reward
         nb.float64, #reward
         nb.float64, #reward_rate
         nb.int32, #reward_trace_length
         nb.float64[:], #decay
         nb.float64[:, :], #trace_history
         nb.int32, #trace_history_length
         nb.int32, #trace_index
         nb.float64[:], #cumulative_activities
         nb.float64[:], #node_activities
         nb.int32[:], #element_index
         nb.int32[:], #parent_index
         nb.int32, #goal_index
         nb.int32, #num_nodes
        ), nopython=True)

def update_rewards(node_reward,
                   reward,
                   reward_rate,
                   trace_length,
                   trace_decay,
                   trace_history,
                   trace_history_length,
                   trace_index,
                   cumulative_activities,
                   node_activities,
                   element_index,
                   parent_index,
                   goal_index,
                   num_nodes
                  ):
    """
    Update node reward estimates.

    Go through the recent history of element goals
    and update the expected reward associated with each
    of them using this time step's reward. The older a goal, the
    less credit it gets for the current reward.

    The parameters are all defined in level.py. It's a runaround,
    but it avoids duplication of documentation.
    """
    # This is a bit of careful bookkeeping.
    # trace_history is a circular (cylindrical) buffer. Generate the
    # appropriate indices for stepping backward in time.
    t_history = np.zeros(trace_length, np.int32)
    for time in range(trace_length):
        # Cycle through the node activity history, starting with the
        # most recent time step and working backward.
        t_past = trace_index - time
        # Once the index reaches the beginning, loop back around.
        if t_past < 0:
            t_past += trace_history_length
        t_history[time] = t_past

    # Now go through each of the nodes. If the node's element is the
    # one that was set as a goal, then it is eligible for updating.
    for i in range(num_nodes):
        # Adjust the reward update rate so that the node adjusts
        # very quickly to the first few exposures and then gradually
        # more slowly to subsequent ones. This tends to decrease training
        # time. It's an agressive learning strategy, and so is
        # prone to initial error, but it allows
        # for later experiences to refine the reward estimate and
        # eventually correct that error out.
        mod_reward_rate = max(reward_rate,
                              1. / (1. + cumulative_activities[i]))
        for time in range(trace_length):
            # Increment the expected reward value associated with each sequence.
            # The size of the increment is larger when:
            #     1. the discrepancy between the previously learned and
            #         observed reward values is larger and
            #     2. the sequence activity is greater.
            # Another way to say this is:
            # If either the reward discrepancy is very small
            # or the sequence activity is very small, there is no change.
            node_reward[i] += ((reward - node_reward[i]) *
                               trace_history[i, t_history[time]] *
                               mod_reward_rate *
                               trace_decay[time])

        # Update the trace history. Write over the oldest record with
        # the newest one.
        parent_activity = node_activities[parent_index[i]]
        if element_index[i] == goal_index:
            trace_history[i, trace_index + 1] = parent_activity
        else:
            trace_history[i, trace_index + 1] = 0.

    # Advance the index of the current time step. Loop around when necessary.
    trace_index += 1
    if trace_index >= trace_history_length:
        trace_index -= trace_history_length

    return node_reward, trace_history, trace_index



"""
The core Becca code

This package contains all the learning algorithms and heuristics
implemented together in a functional learning agent. Most users
won't need to modify anything in this package. Only core developers,
those who want to experiment with and improve the underlying
learning algorithms and their implementations, should need to muck
around with it.

See the becca_test package for examples of how to implement it with a world.
"""



"""
The Level class.
"""

from __future__ import print_function
import numpy as np

import becca.node as node
from becca.ziptie import Ziptie


class Level(object):
    """
    One level in the hierarchy of sequences.

    Each level contains a tree, representing all element sequences of
    length two. The nodes of the tree are an important data structure
    in level-related calculations. Their properties are listed in the
    Level.node_xxx attributes initialized below.

    In each time step, each level takes in new inputs, transforms them
    into sequences, and passes them up to the next level. The level also
    takes sequence goals that are passed down from above, transforms them
    into input goals, and passses them down to the next level. Finally,
    the level uses the new inputs to incrementally train itself for
    future time steps.
    """
    def __init__(self,
                 level_index,
                 max_num_inputs,
                 max_num_elements=None,
                 max_num_sequences=None
                ):
        """
        Configure the level.

        Parameters
        ---------
        level_index : int
            See Level.level_index.
        max_num_inputs : int
            See Level.max_num_inputs.
        max_num_sequences : int
            See Level.max_num_sequences.
        max_num_elements : int
            See Level.max_num_elements.
        """
        # level_index : int
        #     The index of this level.
        #     0 is the first and bottom level, and it counts up from there.
        self.level_index = level_index
        # name : str
        #     The name of this level.
        self.name = "_".join(["level", str(self.level_index)])
        # debug : boolean
        #     Print out extra information about the level's operation.
        self.debug = False

        # max_num_inputs : int
        # max_num_bundles : int
        # max_num_elements : int
        # max_num_nodes : int
        # max_num_sequences : int
        #     The maximum numbers of inputs, bundles, nodes, elements and
        #     sequences and that this level can accept.
        #     max_num_elements = max_num_inputs + max_num_bundles
        self.max_num_inputs = max_num_inputs
        self.max_num_bundles = self.max_num_inputs
        if max_num_elements is None:
            self.max_num_elements = self.max_num_inputs + self.max_num_bundles
        else:
            self.max_num_elements = max_num_elements
        if max_num_sequences is None:
            self.max_num_sequences = self.max_num_inputs
        else:
            self.max_num_sequences = max_num_sequences
        # This limit on the number of nodes accounts for
        # 1 root node, max_num_elements first generation nodes,
        # and each of those has children of all the other elements,
        # max_num_elements - 1.
        # This totals to:
        # 1 + max_num_elements + max_num_elements * (max_num_elements - 1)
        #     = 1 max_num_elements ** 2
        self.max_num_nodes = self.max_num_elements ** 2 + 1
        # num_sequences, num_nodes : int
        #     The number of nodes and sequences that this level
        #     has already created.
        #     num_sequences < num_nodes
        self.num_nodes = 1
        self.num_sequences = 0
        # num_active_elements : int
        #     The number of elements that have exhibited some non-trivial
        #     activity. Initially, the elements corresponding to bundles
        #     will be unused. Until they are populated, it's computationally
        #     more efficient to skip over them.
        self.num_active_elements = self.max_num_inputs
        # Normalization constants.
        # input_max : array of floats
        #     The maximum of each input's activity.
        #     Start with the assumption that each input has a
        #     maximum value of zero. Then adapt up from there
        #     based on the incoming observations.
        # input_max_decay_time, input_max_grow_time : float
        #     The time constant over which maximum estimates are
        #     decreased/increased. Growing time being lower allows the
        #     estimate to be weighted toward the maximum value.
        #self.input_max = np.zeros(self.max_num_inputs)
        self.input_max = np.zeros(self.max_num_inputs)
        self.input_max_grow_time = 1e2
        self.input_max_decay_time = self.input_max_grow_time * 1e2

        # activity_decay_rate : float
        #     The rate at which element activity decays between time steps.
        #     Decay takes five times longer for each additional level.
        self.activity_decay_rate = 1. / (5. ** self.level_index)

        # reward_trace_length : int
        #     The maximum number of time steps between when an action
        #     occurs and when it can claim responsibility for some of
        #     the current reward.
        self.reward_trace_length = 10
        # trace_history_length : int
        #     The length of the action history we need to save
        #     in order to calculate the reward trace.
        #     The extra 2 is due to the implementation of
        #     reward trace calculation.
        self.trace_history_length = self.reward_trace_length + 2
        # decay : array of floats
        #     The discount applied to a reward, depending on how many
        #     time steps after the action it occurred.
        #     if the action was taken at time t,
        #     decay[i] = discount for reward at time step t + 1 + i.
        self.decay = np.zeros(self.reward_trace_length)
        for time in range(self.reward_trace_length):
            # Use a hyperbolic decay.
            # The amount of credit is proportional
            # to 1/t, where t is the number of time steps since the
            # action was taken.
            self.decay[time] = 1. / (time  + 1.)

        # node_trace_history : 2D array of floats
        #     The trace history records, for each time step, all of the
        #     activities of all the nodes that voted for the action that
        #     was taken. These are nodes than can claim some responsibility
        #     for that action, and therefore nodes than can claim some
        #     credit for any subsequent reward. The data structure is
        #     a circular (cylindrical) buffer. Each row in the 2D array
        #     represents a single node. Each column represents a time step.
        #     As time marches forward, node responsibilties are written
        #     in columns to the right. After the last column is written,
        #     the history wraps around, and the next time step is written
        #     onto the first column again. Reading history works the same
        #     way, but working from right to left, wrapping around from
        #     first to last when necessary.
        self.node_trace_history = np.zeros((self.max_num_nodes,
                                            self.trace_history_length))
        # trace_index : int
        #     This is the index of the column in node_trace_history that
        #     represents the current time. It keeps track of
        #     the current location within the buffer.
        self.trace_index = 0

        # input_activities,
        # element_activities,
        # sequence_activitiies : array of floats
        #     inputs, elements, and sequences are all characterized by their
        #     activity--their level of activation at each time step.
        #     Activity for each input, element and sequence
        #     can vary between zero and one. Activity passes upward through
        #     a level. Input activities result in element activities,
        #     which result in sequence activities, which are then passed
        #     up to the next level as its input activities.
        self.input_activities = np.zeros(self.max_num_inputs)
        self.element_activities = np.zeros(self.max_num_elements)
        self.sequence_activities = np.zeros(self.max_num_sequences)

        # input_goals,
        # element_goals,
        # sequence_goals : array of floats
        #     Goals can be set for inputs, elements and sequences.
        #     They are temporary incentives, used for planning and
        #     action selection. They are passed down through the
        #     hierarchy, from higher levels to lower.
        #     Sequence goals are passed down from
        #     the level above. Nodes use those to set element goals.
        #     Element goals are then translated into input goals.
        #     Input goals are passed down to the next level as its
        #     new sequence goals. When the bottom level is reached,
        #     input goals corresponding to actions are used for
        #     action selection.
        self.input_goals = np.zeros(self.max_num_inputs)
        self.element_goals = np.zeros(self.max_num_elements)
        self.sequence_goals = np.zeros(self.max_num_sequences)

        # element_goal_votes : array of floats
        #     Each node votes for an element to become a goal.
        #     Each element keeps the highest vote it receives
        #     in each time step. element_goal_votes tallies
        #     these votes across nodes, and is then reset to
        #     zero at the beginning of eah time step.
        self.element_goal_votes = np.zeros(self.max_num_elements)
        # responsible_nodes : array of ints
        #     This tracks the index of the node that cast the
        #     highest vote for each element. That way, when an
        #     element is selected as a goal, we can see which
        #     node made it happen. This gives the mechanism
        #     some explainability which helps in visualization
        #     and debugging.
        self.responsible_nodes = -np.ones(self.max_num_elements, 'int32')

        # ziptie : Ziptie
        #     The ziptie is an instance of the Ziptie algorithm class,
        #     an incremental method for bundling inputs. Check out
        #     ziptie.py for a complete description. Zipties note which
        #     inputs tend to be co-active and creates bundles of them.
        #     This feature creation mechanism results in l0-sparse
        #     features, which sparsity helps keep Becca fast.
        self.ziptie = Ziptie(self.max_num_inputs,
                             num_bundles=self.max_num_bundles,
                             level=self.level_index)

        # Nodes are the data structure used to represent time.
        # They are arranged into a tree with a single root node,
        # child nodes for each element, and grandchild node for
        # each of the other elements. Each leaf in this tree
        # represents a two-step sequence. Nodes track several
        # aspects of the sequences.

        # node_activities, node_prev_activities : array of floats
        #     The activity of each node at both the current
        #     and the previous time step.
        self.node_activities = np.zeros(self.max_num_nodes)
        self.node_prev_activities = np.zeros(self.max_num_nodes)
        # node_activity_rate : float
        #     The rate at which node activities decay at each time
        #     step in the absence of any additional activation.
        #     new_activity = old_activity * (1 - node_acctivity_rate)
        self.node_activity_rate = self.activity_decay_rate
        # node_cumulative_activities : array of floats
        #     The sum of each node's activity over its lifetime.
        self.node_cumulative_activities = 1e-3 * np.ones(self.max_num_nodes)

        # node_attempts, node_fulfillment, node_unfulfillment : array of floats
        #     An attempt occurs when a node's parent is active and its
        #     terminal element is selected as a goal.
        #     All attempts result in either fulfillment or unfulfillment.
        #     If the node becomes active soon after the attempt, the attempt
        #     is considered fulfilled. The sooner and the larger the activity,
        #     the more complete the fulfillment. All attempts or portions of
        #     attempts that aren't fulfilled get added to unfulfillment.
        self.node_attempts = np.zeros(self.max_num_nodes)
        self.node_fulfillment = 1e-3 * np.ones(self.max_num_nodes)
        self.node_unfulfillment = 1e-3 * np.ones(self.max_num_nodes)

        # node_reward : array of floats
        #     The expected reward of choosing this node's element as a goal
        #     when its parent node is active.
        self.node_reward = np.zeros(self.max_num_nodes)
        # node_reward_rate : float
        #     A constant scaling factor on the rate at which node_reward adapts
        #     to new observations.
        self.node_reward_rate = 3e-3
        # node_curiosity : array of floats
        #     The value associated with selecting a node's element as a goal
        #     when the node's parent is active.
        self.node_curiosity = 1e-1 * np.ones(self.max_num_nodes)
        # node_curiosity_rate : float
        #     A constant scaling factor on the rate at which
        #     curiosity accumulates.
        #     A higher rate yields a more curious and exploratory brain.
        self.node_curiosity_rate = 1e-3

        # node_element_index, node_sequence_index : array of ints
        #     The indices associated with each node's terminal element
        #     and sequence, respectively. Note that nodes initially are
        #     not assigned to sequences. There are far fewer sequences
        #     than nodes and only the most commonly occurring nodes
        #     get elevated to the status of sequence.
        self.node_element_index = -np.ones(self.max_num_nodes, 'int32')
        self.node_sequence_index = -np.ones(self.max_num_nodes, 'int32')
        # node_sequence_threshold : float
        #     If a node's cumulative activity exceeds the
        #     node_sequence_threshold, then the node is elevated to the status
        #     of a sequence, an ouput from one level that becomes an input
        #     to the level above it.
        self.node_sequence_threshold = 1e2 * 4. ** self.level_index
        # node_sequence_length : array of ints
        #     The length of the sequence represented by a node.
        #     For root node, this is 0. For its children, 1.
        #     For the grandchild leaf nodes, it is 2.
        self.node_sequence_length = -np.ones(self.max_num_nodes, 'int32')

        # node_num_children : array of ints
        #     The number of child nodes that belong to each node.
        self.node_num_children = np.zeros(self.max_num_nodes, 'int32')
        # node_child_indices : 2D array of ints
        #     The node indices of the children that belong to each node.
        #     There should be as many non-negative child indices
        #     as their are num_children for each node.
        self.node_child_indices = -np.ones((self.max_num_nodes,
                                            self.max_num_elements), 'int32')
        # node_parent_index : array of ints
        #     The node index of each node's parent node.
        self.node_parent_index = -np.ones(self.max_num_nodes, 'int32')

        # Initialize the node tree.
        # Node 0 is the root. Give it one child for each element.
        # Don't assign sequences to these nodes. They represent sequences
        # of length 1 which aren't interesting. They're already represented
        # in this level's elements.
        self.node_sequence_length[0] = 0
        for i_element in range(self.max_num_elements):
            new_node_index = self.num_nodes
            self.node_child_indices[0, self.node_num_children[0]] = (
                new_node_index)
            self.node_cumulative_activities[new_node_index] = 1.
            self.node_sequence_length[new_node_index] = 1
            self.node_element_index[new_node_index] = i_element
            self.node_parent_index[new_node_index] = 0
            self.node_num_children[0] += 1
            self.num_nodes += 1

            # Give each of these nodes a set of child nodes.
            # This corresponds to initializing all sequences of length 2.
            for j_element in range(self.max_num_elements):
                # Don't allow same-element sequences.
                if i_element != j_element:
                    child_node_index = self.num_nodes
                    self.node_child_indices[
                        new_node_index,
                        self.node_num_children[new_node_index]] = (
                            child_node_index)
                    self.node_cumulative_activities[child_node_index] = 1.
                    self.node_sequence_length[child_node_index] = 2
                    self.node_element_index[child_node_index] = j_element
                    self.node_parent_index[child_node_index] = new_node_index
                    self.node_num_children[new_node_index] += 1
                    self.num_nodes += 1

        self.last_num_nodes = self.num_nodes


    def step(self, new_inputs, reward, satisfaction):
        """
        Advance all the sequences in the level by one time step.

        Start by updating all the inputs. Then walk through all
        the sequences, updating and training.

        Parameters
        ----------
        new_inputs : array of floats
            The inputs collected by the brain for the current time step.
        reward : float
            The current reward value.
        satisfaction : float
            The brain's current state of satisfaction, which is calculated
            from the recent reward history. If it hasn't received much
            reward recently, it won't be very satisfied.
        """
        node_index = 0
        self.update_inputs(new_inputs)

        # Run the inputs through the ziptie to find bundle activities
        # and to learn how to bundle them. nonbundle_activities
        # are all the input activities that don't contribute to
        # bundle_activites. It's what's left over. This forces all the
        # inputs to be expressed as concisely as possible,
        # in bundles wherever possible, rather than separately.
        (nonbundle_activities, bundle_activities) = (
            self.ziptie.featurize(self.input_activities))
        # The element activities are the combination of the residual
        # input activities and the bundle activities.
        self.element_activities = np.concatenate((nonbundle_activities,
                                                  bundle_activities))
        # Incrementally update the bundles in the ziptie.
        self.ziptie.learn()
        self.num_active_elements = self.max_num_inputs + self.ziptie.num_bundles

        # Re-initialize a few quantities for this time step.
        self.responsible_nodes = -np.ones(self.max_num_elements, 'int32')
        self.element_goal_votes = np.zeros(self.max_num_elements)
        prev_parent_activity = 1.
        parent_activity = 1.
        # Kick off the descent down through the node tree.
        self.num_nodes, self.num_sequences = (
            node.step(node_index, # node parameters
                      self.node_activities,
                      self.node_prev_activities,
                      self.node_activity_rate,
                      self.node_cumulative_activities,
                      self.node_attempts,
                      self.node_fulfillment,
                      self.node_unfulfillment,
                      self.node_curiosity_rate,
                      self.node_curiosity,
                      self.node_reward,
                      self.node_element_index,
                      self.node_sequence_index,
                      self.node_sequence_length,
                      self.node_sequence_threshold,
                      self.node_num_children,
                      self.node_child_indices,
                      self.node_parent_index,
                      self.element_activities,
                      self.num_active_elements,
                      parent_activity,
                      prev_parent_activity,
                      self.sequence_activities,
                      self.num_nodes,
                      self.num_sequences,
                      self.max_num_sequences,
                      self.max_num_nodes,
                      self.element_goal_votes,
                      self.responsible_nodes,
                      self.element_goals,
                      self.sequence_goals,
                      reward,
                      satisfaction))

        # If any nodes were added, reflect that.
        if self.num_nodes > self.last_num_nodes:
            # Print any element sequences that have been added this time step.
            if self.debug:
                for i in xrange(self.last_num_nodes, self.num_nodes):
                    sequence_indices = [self.node_element_index[i]]
                    temp_index = i
                    while self.node_parent_index[temp_index] != 0:
                        sequence_indices.append(self.node_element_index[
                            self.node_parent_index[temp_index]])
                        temp_index = self.node_parent_index[temp_index]
                    print('  added sequence', sequence_indices[::-1])
            self.last_num_nodes = self.num_nodes

        # Decide which element goals to select, based on
        # all the votes tallied up across nodes.
        # Choose exactly one goal at each time step.
        # Choose the element with the largest vote. If there is a tie,
        # randomly select between them.
        self.element_goals = np.zeros(self.max_num_elements)
        self.input_goals = np.zeros(self.max_num_inputs)
        matches = np.where(self.element_goal_votes ==
                           np.max(self.element_goal_votes))[0]
        goal_index = matches[np.argmax(np.random.random_sample(matches.size))]
        # Track which node cast the winning vote.
        responsible_node = self.responsible_nodes[goal_index]
        self.element_goals[goal_index] = 1.
        # Express the element goal in terms of input goals.
        if goal_index < self.max_num_inputs:
            self.input_goals[goal_index] = 1.
        else:
            # If the goal element is a bundle, cast it in terms of inputs.
            # Project element goals down to input goals.
            self.input_goals = self.ziptie.get_index_projection(
                goal_index - self.max_num_inputs)

        if self.debug:
            print('element goal votes', self.element_goal_votes)
            print('element goals', self.element_goals)
            print('input goals', self.input_goals)
            self.print_node(responsible_node)

        # Assign credit for the current reward to any goals set
        # and actions taken in the last few time steps.
        self.node_reward, self.node_trace_history, self.trace_index = (
            node.update_rewards(self.node_reward,
                                reward,
                                self.node_reward_rate,
                                self.reward_trace_length,
                                self.decay,
                                self.node_trace_history,
                                self.trace_history_length,
                                self.trace_index,
                                self.node_cumulative_activities,
                                self.node_activities,
                                self.node_element_index,
                                self.node_parent_index,
                                goal_index,
                                self.num_nodes
                               ))
        # If any of the nodes are assigned to this level's
        # sequences, their activities will be reflected
        # in the sequence activities and pushed upward to be
        # inputs for the next level.
        return self.sequence_activities


    def update_inputs(self, inputs, start_index=0):
        """
        Normalize and update inputs.

        Normalize activities so that they are predictably distrbuted.
        Use a running estimate of the maximum of each cable activity.
        Scale it so that the max would fall at 1.

        Normalization has several benefits.
        1. It makes for fewer constraints on worlds and sensors.
           It allows any sensor can return any range of values.
        2. Gradual changes in sensors and the world can be adapted to.
        3. It makes the bundle creation heuristic more robust and
           predictable. The approximate distribution of cable
           activities is known and can be designed for.

        After normalization, update each input activity with either
        a) the normalized value of its corresponding input or
        b) the decayed value, carried over from the previous time step,
        whichever is greater.

        Parameters
        ----------
        inputs : array of floats
            The current activity of the inputs.
        start_index : int
            The first input to update.

        Returns
        -------
        None
            self.input_activities is modified to include
            the normalized values of each of the inputs.
        """
        epsilon = 1e-8

        if start_index + inputs.size > self.max_num_inputs:
            print("level.Level.update_inputs:")
            print("    Attempting to update out of range input activities.")

        # This is written to be easily compilable by numba, however,
        # it hasn't proven to be at all significant in profiling, so it
        # has stayed in slow-loop python for now.
        stop_index = min(start_index + inputs.size, self.max_num_inputs)
        # Input index
        j = 0
        for i in range(start_index, stop_index):
            val = inputs[j]

            # Decay the maximum value.
            self.input_max[i] += ((val - self.input_max[i]) /
                                  self.input_max_decay_time)

            # Grow the maximum value, when appropriate.
            if val > self.input_max[i]:
                self.input_max[i] += ((val - self.input_max[i]) /
                                      self.input_max_grow_time)

            # Scale the input by the maximum.
            val = val / (self.input_max[i] + epsilon)
            # Ensure that 0 <= val <= 1.
            val = max(0., val)
            val = min(1., val)

            self.input_activities[i] = max(val,
                                           self.input_activities[i] *
                                           (1. - self.activity_decay_rate))
            j += 1


    def visualize(self):
        """
        Show the current state of the level.
        """
        # activity_threshold the level at which we can ignore
        # an element's activity in order to simplify display.
        activity_threshold = .01
        print(self.name)

        print("Input activities")
        for i_input, activity in enumerate(self.input_activities):
            if activity > activity_threshold:
                print(" ".join(["input", str(i_input), ":",
                                "activity ", str(activity)]))

        print("Sequence activities")
        for i_sequence, activity in enumerate(self.sequence_activities):
            if activity > activity_threshold:
                print(" ".join(["sequence", str(i_sequence), ":",
                                "activity ", str(activity)]))

        # Enumerate all sequences, doing a depth-first descent
        # through the tree.
        def descend(node_index,
                    sequence_nodes):
            """
            Recursively descend the node trees and enumerate the sequences.
            """
            # If the current node index has a corresponding sequence,
            # add it to the list.
            sequence_nodes.append(node_index)

            # If this is a terminal node, backtrack up the tree.
            if self.node_num_children[node_index] == 0:
                return

            # If this isn't a terminal node, descend through all the
            # child children.
            for child_index in self.node_child_indices[
                    node_index, :self.node_num_children[node_index]]:
                if (self.node_element_index[child_index] <
                        self.num_active_elements):
                    descend(child_index, sequence_nodes)

        root_index = 0
        sequence_nodes = []
        descend(root_index,
                sequence_nodes)
        for i in sequence_nodes:
            self.print_node(i)
        print("==============================================================")


    def print_node(self, i):
        """
        Print a bunch of information about a node.
        """
        print("----------------------------------------------")

        # Re-create the node's sequence by tracing its parentage.
        full_sequence = [self.node_element_index[i]]
        temp_index = i
        while self.node_parent_index[temp_index] > 0:
            full_sequence.append(self.node_element_index[
                self.node_parent_index[temp_index]])
            temp_index = self.node_parent_index[temp_index]
        full_sequence = full_sequence[::-1]
        sequence_index = self.node_sequence_index[i]
        if sequence_index == -1:
            sequence_index = "none"
        print("  node", i,
              "  sequence", sequence_index,
              ": ", full_sequence)

        #Show the transitions within the sequence.
        print("    cumulative: {0:.4f}".format(
            self.node_cumulative_activities[i]))
        print("    curiosity: {0:.4f}".format(self.node_curiosity[i]))
        print("    reward: {0:.4f}".format(self.node_reward[i]))
        total_goal_value = self.node_reward[i] + self.node_curiosity[i]
        print("    total goal value: {0:.4f}".format(total_goal_value))



"""
A set of functions written to help zipties take advantage of numba.

After a little experimentation, it appears that numba works best
when everything is written out explicitly in loops. It even beats
numpy handily on dense matrices. It doesn't like lists or lists of lists.
It also tends to work better if you
don't use numpy functions and if you make your code as simple
as possible.

The (nopython=True) call makes it so that if numba can't compile the code
to C (very fast), but is forced to fall back to python instead (dead slow
when doing loops), the function will fail and throw an error.
"""
from numba import jit


@jit(nopython=True)
def set_dense_val(array2d, i_rows, i_cols, val):
    """
    Set values in a dense 2D array using a list of indices.

    Parameters
    ----------
    array2d : 2D array of floats
        The array in which to set values.
    i_rows, i_cols: array of ints
        The row and column indices of each element to change.
    val : float
        The new value to assign.

    Returns
    -------
    Occur indirectly by modifying array2d.
    """
    for i, _ in enumerate(i_rows):
        array2d[i_rows[i], i_cols[i]] = val


@jit(nopython=True)
def max_dense(array2d, results):
    """
    Find the maximum value of a dense 2D array, with its row and column

    Parameters
    ----------
    array2d : 2D array of floats
        The array to find the maximum value of.
    results : array of floats, size 3
        An array for holding the results of the operation.

    Returns
    -------
    Results are returned indirectly by modifying results.
    The results array has three elements and holds
        [0] the maximum value found
        [1] the row number in which it was found
        [2] the column number in which it was found
    """
    max_val = results[0]
    i_row_max = results[1]
    i_col_max = results[2]
    for i_row in range(array2d.shape[0]):
        for i_col in range(array2d.shape[1]):
            if array2d[i_row, i_col] > max_val:
                max_val = array2d[i_row, i_col]
                i_row_max = i_row
                i_col_max = i_col
    results[0] = max_val
    results[1] = i_row_max
    results[2] = i_col_max


@jit(nopython=True)
def find_bundle_activities(i_rows, i_cols, cables, bundles, weights, threshold):
    """
    Use a greedy method to sparsely translate cables to bundles.

    Start at the last bundle added and work backward to the first.
    For each, calculate the bundle activity by finding the minimum
    value of each of its constituent cables. Then subtract out
    the bundle activity from each of its cables.

    Parameters
    ----------
    bundles : 1D array of floats
        An array of bundle activity values. Initially it is all zeros.
    cables : 1D array of floats
        An array of cable activity values.
    i_rows : array of ints
        The row indices for the non-zero sparse 2D array..
    i_cols : array of ints
        The column indices for the non-zero sparse 2D array elements.
        All non-zero elements are assumed to be 1.
        i_rows and i_cols must be the same length.
        Each column represents a cable and each row represents a bundle.
        The 2D array is a map from cables to bundles.
    threshold : float
        The amount of bundle activity below which, we just don't care.
    weights : array of floats
        A multiplier for how strongly the activity of each bundle should
        be considered when greedily selecting the next one to activate.

    Results
    -------
    Returned indirectly by modifying `cables. These are the residual
    cable activities that are not represented by any bundle activities.
    """
    large = 1e10
    max_vote = large

    # Repeat this process until the residual cable activities don't match
    # any bundles well.
    while max_vote > threshold:
        # Initialize the loop that greedily looks for the most strongly
        # activated bundle.
        max_vote = 0.
        best_val = 0.
        best_bundle = 0
        # This is the index in i_row and i_col where the
        # current bundle's cable constituents are listed. Cable indices
        # are assumed to be contiguous and bundles are assumed to be
        # listed in ascending order of index.
        i_best_bundle = 0

        # Iterate over each bundle, that is, over each row.
        i = len(i_rows) - 1
        row = i_rows[i]
        while row > -1:

            # For each bundle, find the minimum cable activity that
            # contribues to it.
            min_val = large
            n_cables = 0.
            i_bundle = i
            while i_rows[i] == row and i > -1:
                col = i_cols[i]
                val = cables[col]
                if val < min_val:
                    min_val = val
                n_cables += 1.
                i -= 1

            # The strength of the vote for the bundle is the minimum cable
            # activity multiplied by the number of cables. This weights
            # bundles with many member cables more highly than bundles
            # with few cables. It is a way to encourage sparsity and to
            # avoid creating more bundles than necessary.
            vote = min_val * (1. + .1 * (n_cables - 1.)) * (1. + weights[row])

            # Update the winning bundle if appropriate.
            if vote > max_vote:
                max_vote = vote
                best_val = min_val
                i_best_bundle = i_bundle
                best_bundle = row

            # Move on to the next bundle.
            row -= 1

        if best_val > 0.:
            # Set the bundle activity.
            bundles[best_bundle] = best_val

            # Subtract the bundle activity from each of the cables.
            # Using i_best_bundle lets us jump right to the place in
            # the list of indices where the cables for the winning bundle
            # are listed.
            i = i_best_bundle
            while i_rows[i] == best_bundle and i > -1:
                col = i_cols[i]
                cables[col] -= best_val
                i -= 1


@jit(nopython=True)
def nucleation_energy_gather(nonbundle_activities, nucleation_energy):
    """
    Gather nucleation energy.

    This formulation takes advantage of loops and the sparsity of the data.
    The original arithmetic looks like
        nucleation_energy += (nonbundle_activities *
                              nonbundle_activities.T *
                              nucleation_energy_rate)
    Parameters
    ----------
    nonbundle_activities : array of floats
        The current activity of each input feature that is not explained by
        or captured in a bundle.
    nucleation_energy : 2D array of floats
        The amount of nucleation energy accumulated between each pair of
        input features.

    Results
    -------
    Returned indirectly by modifying nucleation_energy.
    """
    for i_cable1, _ in enumerate(nonbundle_activities):
        activity1 = nonbundle_activities[i_cable1]
        if activity1 > 0.:
            for i_cable2, _ in enumerate(nonbundle_activities):
                activity2 = nonbundle_activities[i_cable2]
                if activity2 > 0.:
                    nucleation_energy[i_cable1, i_cable2] += (
                        activity1 * activity2)


@jit(nopython=True)
def agglomeration_energy_gather(bundle_activities, nonbundle_activities,
                                n_bundles, agglomeration_energy):
    """
    Accumulate the energy binding a new feature to an existing bundle..

    This formulation takes advantage of loops and the sparsity of the data.
    The original arithmetic looks like
        coactivities = bundle_activities * nonbundle_activities.T
        agglomeration_energy += coactivities * agglomeration_energy_rate

    Parameters
    ----------
    bundle_activities : array of floats
        The activity level of each bundle.
    nonbundle_activities : array of floats
        The current activity of each input feature that is not explained by
        or captured in a bundle.
    n_bundles : int
        The number of bundles that have been created so far.
    agglomeration_energy : 2D array of floats
        The total energy that has been accumulated between each input feature
        and each bundle.

    Results
    -------
    Returned indirectly by modifying `agglomeration_energy.
    """
    for i_col, _ in enumerate(nonbundle_activities):
        activity = nonbundle_activities[i_col]
        if activity > 0.:
            # Only decay bundles that have been created
            for i_row in range(n_bundles):
                if bundle_activities[i_row] > 0.:
                    coactivity = activity * bundle_activities[i_row]
                    agglomeration_energy[i_row, i_col] += coactivity



"""
Connect a world with a brain and set it to running.
"""
import numpy as np

from becca.brain import Brain

def run(world, restore=False):
    """
    Run Becca with a world.

    Connect the brain and the world together and runs them for as long
    as the world dictates.

    Parameters
    ----------
    restore : bool, optional
        If ``restore`` is True, try to restore the brain
        from a previously saved
        version, picking up where it left off.
        Otherwise it create a new one. The default is False.

    Returns
    -------
    performance : float
        The performance of the brain over its lifespan, measured by the
        average reward it gathered per time step.
    """
    brain_name = '{0}_brain'.format(world.name)
    brain = Brain(world.num_sensors,
                  world.num_actions,
                  brain_name=brain_name)
    if restore:
        brain = brain.restore()
    # Start at a resting state.
    actions = np.zeros(world.num_actions)
    sensors, reward = world.step(actions)
    # Repeat the loop through the duration of the existence of the world:
    # sense, act, repeat.
    while world.is_alive():
        actions = brain.sense_act_learn(sensors, reward)
        sensors, reward = world.step(actions)
        world.visualize(brain)
    # Wrap up the run.
    performance = brain.report_performance()
    return performance



"""
The Ziptie class.
"""

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt

import becca.tools as tools
import becca.ziptie_numba as nb


class Ziptie(object):
    """
    An incremental unsupervised clustering algorithm.

    Input channels are clustered together into mutually co-active sets.
    A helpful metaphor is bundling cables together with zip ties.
    Cables that carry related signals are commonly co-active,
    and are grouped together. Cables that are co-active with existing
    bundles can be added to those bundles. A single cable may be ziptied
    into several different bundles. Co-activity is estimated
    incrementally, that is, the algorithm updates the estimate after
    each new set of signals is received.

    Zipties are arranged hierarchically within the agent's drivetrain.
    The agent begins with only one ziptie and creates subsequent
    zipties as previous ones mature.

    When stacked with other levels,
    zipties form a sparse deep neural network (DNN).
    This DNN has the extremely desirable characteristic of
    l-0 sparsity--the number of non-zero weights are minimized.
    The vast majority of weights in this network are zero,
    and the rest are one.
    This makes sparse computation feasible and allows for
    straightforward interpretation and visualization of the
    features at each level.
    """


    def __init__(self,
                 num_cables,
                 num_bundles=None,
                 level=0,
                 name=None,
                 debug=False):
        """
        Initialize each level, pre-allocating data structures.

        Parameters
        ----------
        debug : boolean
            Indicate whether to print informative status messages
            during execution.
        level : int
            The position of this Ziptie in the hierarchy.
        num_bundles : int
            The number of bundle outputs from the Ziptie.
        num_cables : int
            The number of inputs to the Ziptie.
        name : str
            The name assigned to the Ziptie.
            Default is 'anonymous'.
        """
        # level : int
        #     The position of this Ziptie in the hierarchy.
        self.level = level
        # name : str
        #     The name associated with the Ziptie.
        if name is None:
            self.name = '_'.join(['ziptie', str(self.level)])
        else:
            self.name = name
        # debug : boolean
        #     Indicate whether to print informative status messages
        #     during execution.
        self.debug = debug
        # max_num_cables : int
        #     The maximum number of cable inputs allowed.
        self.max_num_cables = num_cables
        # max_num_bundles : int
        #     The maximum number of bundle outputs allowed.
        if num_bundles is None:
            self.max_num_bundles = self.max_num_cables
        else:
            self.max_num_bundles = num_bundles
        # num_bundles : int
        #     The number of bundles that have been created so far.
        self.num_bundles = 0
        # nucleation_threshold : float
        #     Threshold above which nucleation energy results in nucleation.
        #     With each level, increase this by a factor of 5. This gives
        #     Each level a chance to mature before the one above it.
        self.nucleation_threshold = 50. * 5. ** self.level
        # agglomeration_threshold
        #     Threshold above which agglomeration energy results
        #     in agglomeration.
        self.agglomeration_threshold = .5 * self.nucleation_threshold
        # activity_threshold : float
        #     Threshold below which input activity is teated as zero.
        #     By ignoring the small activity values,
        #     computation gets much faster.
        self.activity_threshold = .1
        # bundles_full : bool
        #     If True, all the bundles in the Ziptie are full
        #     and learning stops. This is another way to speed up
        #     computation.
        self.bundles_full = False
        # cable_activities : array of floats
        #     The current set of input actvities.
        self.cable_activities = np.zeros(self.max_num_cables)
        # bundle_activities : array of floats
        #     The current set of bundle activities.
        self.bundle_activities = np.zeros(self.max_num_bundles)
        # nonbundle_activities : array of floats
        #     The set of input activities that do not contribute
        #     to any of the current bundle activities.
        self.nonbundle_activities = np.zeros(self.max_num_cables)

        # bundle_map_size : int
        #     The maximum number of non-zero entries in the bundle map.
        self.bundle_map_size = 8
        # bundle_map_cols, bundle_map_rows : array of ints
        #     To represent the sparse 2D bundle map, a pair of row and col
        #     arrays are used. Rows are bundle indices, and cols are
        #     feature indices.  The bundle map shows which cables
        #     are zipped together to form which bundles.
        self.bundle_map_rows = -np.ones(self.bundle_map_size).astype(int)
        self.bundle_map_cols = -np.ones(self.bundle_map_size).astype(int)
        # n_map_entries: int
        #     The total number of bundle map entries that
        #     have been created so far.
        self.n_map_entries = 0
        # agglomeration_energy : 2D array of floats
        #     The accumulated agglomeration energy for each bundle-cable pair.
        self.agglomeration_energy = np.zeros((self.max_num_bundles,
                                              self.max_num_cables))
        # nucleation_energy : 2D array of floats
        #     The accumualted nucleation energy associated
        #     with each cable-cable pair.
        self.nucleation_energy = np.zeros((self.max_num_cables,
                                           self.max_num_cables))


    def featurize(self, new_cable_activities, bundle_weights=None):
        """
        Calculate how much the cables' activities contribute to each bundle.

        Find bundle activities by taking the minimum input value
        in the set of cables in the bundle. The bulk of the computation
        occurs in ziptie_numba.find_bundle_activities.
        """
        self.nonbundle_activities = new_cable_activities.copy()
        self.bundle_activities = np.zeros(self.max_num_bundles)
        if bundle_weights is None:
            bundle_weights = np.ones(self.max_num_bundles)
        if self.n_map_entries > 0:
            nb.find_bundle_activities(
                self.bundle_map_rows[:self.n_map_entries],
                self.bundle_map_cols[:self.n_map_entries],
                self.nonbundle_activities,
                self.bundle_activities,
                bundle_weights, self.activity_threshold)
        # The residual cable_activities after calculating
        # bundle_activities are the nonbundle_activities.
        # Sparsify them by setting all the small values to zero.
        self.nonbundle_activities[np.where(self.nonbundle_activities <
                                           self.activity_threshold)] = 0.
        return self.nonbundle_activities, self.bundle_activities


    def learn(self):
        """
        Update co-activity estimates and calculate bundle activity

        This step combines the projection of cables activities
        to bundle activities together with using the cable activities
        to incrementally train the Ziptie.

        Parameters
        ----------
        none

        Returns
        -------
        none
        """
        if not self.bundles_full:
            self._create_new_bundles()
            self._grow_bundles()
        return


    def _create_new_bundles(self):
        """
        If the right conditions have been reached, create a new bundle.
        """
        # Incrementally accumulate nucleation energy.
        nb.nucleation_energy_gather(self.nonbundle_activities,
                                    self.nucleation_energy)

        # Don't accumulate nucleation energy between a cable and itself
        ind = np.arange(self.cable_activities.size).astype(int)
        self.nucleation_energy[ind, ind] = 0.

        # Don't accumulate nucleation energy between cables already
        # in the same bundle
        for i in range(self.n_map_entries):
            i_bundle = self.bundle_map_rows[i]
            i_cable = self.bundle_map_cols[i]
            j = 1
            j_bundle = self.bundle_map_rows[i + j]
            j_cable = self.bundle_map_cols[i + j]
            while j_bundle == i_bundle:
                self.nucleation_energy[i_cable, j_cable] = 0.
                self.nucleation_energy[j_cable, i_cable] = 0.
                j += 1
                j_bundle = self.bundle_map_rows[i + j]
                j_cable = self.bundle_map_cols[i + j]

        results = -np.ones(3)
        nb.max_dense(self.nucleation_energy, results)
        max_energy = results[0]
        cable_index_a = int(results[1])
        cable_index_b = int(results[2])

        # Add a new bundle if appropriate
        if max_energy > self.nucleation_threshold:
            self.bundle_map_rows[self.n_map_entries] = self.num_bundles
            self.bundle_map_cols[self.n_map_entries] = cable_index_a
            self.increment_n_map_entries()
            self.bundle_map_rows[self.n_map_entries] = self.num_bundles
            self.bundle_map_cols[self.n_map_entries] = cable_index_b
            self.increment_n_map_entries()
            self.num_bundles += 1

            print(' '.join(['    ', self.name,
                            'bundle', str(self.num_bundles),
                            'added with cables', str(cable_index_a),
                            str(cable_index_b)]))

            # Check whether the Ziptie's capacity has been reached.
            if self.num_bundles == self.max_num_bundles:
                self.bundles_full = True

            # Reset the accumulated nucleation and agglomeration energy
            # for the two cables involved.
            self.nucleation_energy[cable_index_a, :] = 0.
            self.nucleation_energy[cable_index_b, :] = 0.
            self.nucleation_energy[:, cable_index_a] = 0.
            self.nucleation_energy[:, cable_index_b] = 0.
            self.agglomeration_energy[:, cable_index_a] = 0.
            self.agglomeration_energy[:, cable_index_b] = 0.


    def _grow_bundles(self):
        """
        Update an estimate of co-activity between all cables.
        """
        # Incrementally accumulate agglomeration energy.
        nb.agglomeration_energy_gather(self.bundle_activities,
                                       self.nonbundle_activities,
                                       self.num_bundles,
                                       self.agglomeration_energy)

        # Don't accumulate agglomeration energy between cables already
        # in the same bundle
        val = 0.
        if self.n_map_entries > 0:
            nb.set_dense_val(self.agglomeration_energy,
                             self.bundle_map_rows[:self.n_map_entries],
                             self.bundle_map_cols[:self.n_map_entries],
                             val)

        results = -np.ones(3)
        nb.max_dense(self.agglomeration_energy, results)
        max_energy = results[0]
        cable_index = int(results[2])
        bundle_index = int(results[1])

        # Add a new bundle if appropriate
        if max_energy > self.agglomeration_threshold:
            # Find which cables are in the new bundle.
            cables = [cable_index]
            for i in range(self.n_map_entries):
                if self.bundle_map_rows[i] == bundle_index:
                    cables.append(self.bundle_map_cols[i])
            # Check whether the agglomeration is already in the bundle map.
            candidate_bundles = np.arange(self.num_bundles)
            for cable in cables:
                matches = np.where(self.bundle_map_cols == cable)[0]
                candidate_bundles = np.intersect1d(
                    candidate_bundles,
                    self.bundle_map_rows[matches],
                    assume_unique=True)
            if candidate_bundles.size != 0:
                # The agglomeration has already been used to create a
                # bundle. Ignore and reset they count. This can happen
                # under normal circumstances, because of how nonbundle
                # activities are calculated.
                self.agglomeration_energy[bundle_index, cable_index] = 0.
                return

            # Make a copy of the growing bundle.
            for i in range(self.n_map_entries):
                if self.bundle_map_rows[i] == bundle_index:
                    self.bundle_map_rows[self.n_map_entries] = self.num_bundles
                    self.bundle_map_cols[self.n_map_entries] = (
                        self.bundle_map_cols[i])
                    self.increment_n_map_entries()
            # Add in the new cable.
            self.bundle_map_rows[self.n_map_entries] = self.num_bundles
            self.bundle_map_cols[self.n_map_entries] = cable_index
            self.increment_n_map_entries()
            self.num_bundles += 1

            if self.debug:
                print(' '.join(['    ', self.name,
                                'bundle', str(self.num_bundles),
                                'added: bundle', str(bundle_index),
                                'and cable', str(cable_index)]))

            # Check whether the Ziptie's capacity has been reached.
            if self.num_bundles == self.max_num_bundles:
                self.bundles_full = True

            # Reset the accumulated nucleation and agglomeration energy
            # for the two cables involved.
            self.nucleation_energy[cable_index, :] = 0.
            self.nucleation_energy[cable_index, :] = 0.
            self.nucleation_energy[:, cable_index] = 0.
            self.nucleation_energy[:, cable_index] = 0.
            self.agglomeration_energy[:, cable_index] = 0.
            self.agglomeration_energy[bundle_index, :] = 0.


    def increment_n_map_entries(self):
        """
        Add one to n_map entries and grow the bundle map as needed.
        """
        self.n_map_entries += 1
        if self.n_map_entries >= self.bundle_map_size:
            self.bundle_map_size *= 2
            self.bundle_map_rows = tools.pad(self.bundle_map_rows,
                                             self.bundle_map_size,
                                             val=-1, dtype='int')
            self.bundle_map_cols = tools.pad(self.bundle_map_cols,
                                             self.bundle_map_size,
                                             val=-1, dtype='int')


    def get_index_projection(self, bundle_index):
        """
        Project bundle_index down to its cable indices.

        Parameters
        ----------
        bundle_index : int
            The index of the bundle to project onto its constituent cables.

        Returns
        -------
        projection : array of floats
            An array of zeros and ones, representing all the cables that
            contribute to the bundle. The values projection
            corresponding to all the cables that contribute are 1.
        """
        projection = np.zeros(self.max_num_cables)
        for i in range(self.n_map_entries):
            if self.bundle_map_rows[i] == bundle_index:
                projection[self.bundle_map_cols[i]] = 1.
        return projection


    def visualize(self):
        """
        Turn the state of the Ziptie into an image.
        """
        print(' '.join(['ziptie', str(self.level)]))
        # First list the bundles andthe cables in each.
        i_bundles = self.bundle_map_rows[:self.n_map_entries]
        i_cables = self.bundle_map_cols[:self.n_map_entries]
        i_bundles_unique = np.unique(i_bundles)
        if i_bundles_unique is not None:
            for i_bundle in i_bundles_unique:
                b_cables = list(np.sort(i_cables[np.where(
                    i_bundles == i_bundle)[0]]))
                print(' '.join(['    bundle', str(i_bundle),
                                'cables:', str(b_cables)]))

        plot = False
        if plot:
            if self.n_map_entries > 0:
                # Render the bundle map.
                bundle_map = np.zeros((self.max_num_cables,
                                       self.max_num_bundles))
                nb.set_dense_val(bundle_map,
                                 self.bundle_map_rows[:self.n_map_entries],
                                 self.bundle_map_cols[:self.n_map_entries], 1.)
                tools.visualize_array(bundle_map,
                                      label=self.name + '_bundle_map')

                # Render the agglomeration energy.
                label = '_'.join([self.name, 'agg_energy'])
                tools.visualize_array(self.agglomeration_energy, label=label)
                plt.xlabel(str(np.max(self.agglomeration_energy)))

                # Render the nucleation energy.
                label = '_'.join([self.name, 'nuc_energy'])
                tools.visualize_array(self.nucleation_energy, label=label)
                plt.xlabel(str(np.max(self.nucleation_energy)))



from setuptools import setup

setup(name='becca',
version='0.8.0',
description='A general learning algorithm',
url='http://github.com/brohrer/becca',
download_url='https://github.com/brohrer/becca/archive/master.zip',
author='Brandon Rohrer',
author_email='brohrer@gmail.com',
license='MIT',
packages=['becca'],
zip_safe=False)



"""
The Brain class.
"""

from __future__ import print_function
import cPickle as pickle
import os
import numpy as np

from becca.affect import Affect
from becca.level import Level


class Brain(object):
    """
    A biologically motivated learning algorithm.

    Becca's Brain contains all of its learning algorithms,
    integrated into a single whole.
    Check out connector.py for an example for how to attach a world
    to a brain.
    """
    def __init__(self, num_sensors, num_actions, brain_name='test_brain'):
        """
        Configure the Brain.

        Parameters
        ----------
        brain_name : str
            A descriptive string identifying the brain.
        num_actions : array of ints
            The total number of action outputs that the world is expecting.
        num_sensors : array of ints
            The total number of sensor inputs that the world is providing.
        """
        # num_sensors : int
        #     The number of distinct sensors that the world will be passing in
        #     to the brain.
        self.num_sensors = num_sensors
        # num_actions : int
        #     The number of distinct actions that the brain can choose to
        #     execute in the world. Always include an extra action.
        #     The last is the 'do nothing' action.
        self.num_actions = num_actions + 1

        # backup_interval : int
        #     The number of time steps between saving a copy of the brain
        #     out to a pickle file for easy recovery.
        self.backup_interval = 1e5
        # name : str
        #     Unique name for this brain.
        self.name = brain_name

        # Identify the full local path of the brain.py module.
        # This trick is used to conveniently locate other Becca resources.
        module_path = os.path.dirname(os.path.abspath(__file__))
        # log_dir : str
        #     Relative path to the log directory. This is where backups
        #     and images of the brain's state and performance are kept.
        self.log_dir = os.path.normpath(os.path.join(module_path, 'log'))
        # Check whether the directory is already there. If not, create it.
        if not os.path.isdir(self.log_dir):
            os.makedirs(self.log_dir)
        # pickle_filename : str
        #     Relative path and filename of the backup pickle file.
        self.pickle_filename = os.path.join(self.log_dir,
                                            '{0}.pickle'.format(brain_name))
        # affect : Affect
        #     See the pydocs in the module affect.py for the class Affect.
        self.affect = Affect()
        # satisfaction : float
        #     The level of contentment experienced by the brain.
        #     Higher contentment dampens curiosity and the drive to explore.
        self.satisfaction = 0.

        # levels : list of Level
        #     Collectively, the levels form a hierarchy with levels[0]
        #     on the bottom.
        #     Refer to level.py for a detailed description of a level.
        #     Initialize the 0th level.
        num_inputs = self.num_sensors + self.num_actions
        level_index = 0
        level_0 = Level(level_index, num_inputs)
        self.levels = [level_0]

        # actions : array of floats
        #     The set of actions to execute this time step.
        self.actions = np.zeros(self.num_actions)

        # timestep : int
        #     The age of the brain in discrete time steps.
        self.timestep = 0


    def sense_act_learn(self, sensors, reward):
        """
        Take sensor and reward data in and use them to choose an action.

        Parameters
        ----------
        sensors : array of floats
            The information coming from the sensors in the world.
            The array should have self.num_sensors inputs.
            Each value in the array is expected to be between 0 and 1,
            inclusive. Sensor values are interpreted as fuzzy binary
            values, rather than continuous values. For instance,
            the brain doesn't interpret a contact sensor value of .5
            to mean that the contact
            sensor was only weakly contacted. It interprets it
            to mean that the sensor was fully contacted for 50% of the sensing
            duration or that there is a 50% chance that the sensor was
            fully contacted during the entire sensing duration. For another
            example, a light sensor reading of zero won't be
            interpreted as by the brain as darkness. It will just be
            interpreted as a lack of information about the lightness.
        reward : float
            The extent to which the brain is being rewarded by the
            world. It is expected to be between -1 and 1, inclusive.
            -1 is the worst pain ever. 1 is the most intense ecstasy
            imaginable. 0 is neutral.

        Returns
        -------
        actions : array of floats
            The action commands that the brain is sending to the world
            to be executed. The array should have self.num_actions
            inputs in it. Each value should be binary: 0 and 1. This
            allows the brain to learn most effectively how to interact
            with the world to obtain more reward.
        """
        self.timestep += 1

        # Calculate the "mood" of the agent.
        self.satisfaction = self.affect.update(reward)

        # Calcuate activities of all the sequences in the hierarchy.
        input_activities = np.concatenate((sensors, self.actions))
        for level in self.levels:
            sequence_activities = level.step(input_activities,
                                             reward,
                                             self.satisfaction)
            # For the next level
            input_activities = sequence_activities

        # If the top level has more than half of its allocated sequences
        # created, then create a new level on top of it.
        level = self.levels[-1]
        if level.num_sequences > level.max_num_sequences / 2.:
            # Initialize the next level.
            num_inputs = level.max_num_sequences
            level_index = level.level_index + 1
            print('------------------------------------------ Creating level',
                  level_index)
            next_level = Level(level_index, num_inputs)
            self.levels.append(next_level)

            sequence_activities = next_level.step(input_activities,
                                                  reward,
                                                  self.satisfaction)
        # Pass goals back down.
        for i in range(len(self.levels) - 1)[::-1]:
            self.levels[i].sequence_goals = self.levels[i + 1].input_goals

        # Isolate the actions from the rest of the goals..
        self.actions = self.levels[0].input_goals[
            self.num_sensors:self.num_sensors + self.num_actions]

        # debug: Random actions
        #self.actions = self.random_actions()

        # Periodically back up the brain.
        if (self.timestep % self.backup_interval) == 0:
            self.backup()

        # Account for the fact that the last "do nothing" action
        # was added by the brain.
        return self.actions[:-1]


    def random_actions(self):
        """
        Generate a random set of actions.

        This is used for debugging. Running a world with random
        actions gives a baseline performance floor.

        Returns
        -------
        actions : array of floats
            See sense_act_learn.actions.
        """
        threshold = .5 / float(self.num_actions)
        action_strength = np.random.random_sample(self.num_actions)
        actions = np.zeros(self.num_actions)
        actions[np.where(action_strength < threshold)] = 1.
        return actions


    def report_performance(self):
        """
        Make a report of how the brain did over its lifetime.

        Returns
        -------
        performance : float
            The average reward per time step collected by
            the brain over its lifetime.
        """
        return self.affect.visualize(self.timestep, self.name, self.log_dir)


    def backup(self):
        """
        Archive a copy of the brain object for future use.

        Returns
        -------
        success : bool
            If the backup process completed without any problems, success
            is True, otherwise it is False.
        """
        success = False
        try:
            with open(self.pickle_filename, 'wb') as brain_data:
                pickle.dump(self, brain_data)
            # Save a second copy. If you only save one, and the user
            # happens to ^C out of the program while it is being saved,
            # the file becomes corrupted, and all the learning that the
            # brain did is lost.
            with open('{0}.bak'.format(self.pickle_filename),
                      'wb') as brain_data_bak:
                pickle.dump(self, brain_data_bak)
        except IOError as err:
            print('File error: {0} encountered while saving brain data'.
                  format(err))
        except pickle.PickleError as perr:
            print('Pickling error: {0} encountered while saving brain data'.
                  format(perr))
        else:
            success = True
        return success


    def restore(self):
        """
        Reconstitute the brain from a previously saved brain.

        Returns
        -------
        restored_brain : Brain
            If restoration was successful, the saved brain is returned.
            Otherwise a notification prints and a new brain is returned.
        """
        restored_brain = self
        try:
            with open(self.pickle_filename, 'rb') as brain_data:
                loaded_brain = pickle.load(brain_data)

            # Compare the number of channels in the restored brain with
            # those in the already initialized brain. If it matches,
            # accept the brain. If it doesn't,
            # print a message, and keep the just-initialized brain.
            # Sometimes the pickle file is corrputed. When this is the case
            # you can manually overwrite it by removing the .bak from the
            # .pickle.bak file. Then you can restore from the backup pickle.
            if ((loaded_brain.num_sensors == self.num_sensors) and
                    (loaded_brain.num_actions == self.num_actions)):
                print('Brain restored at timestep {0} from {1}'.format(
                    str(loaded_brain.timestep), self.pickle_filename))
                restored_brain = loaded_brain
            else:
                print('The brain {0} does not have the same number'.format(
                    self.pickle_filename))
                print('of sensors and actions as the world.')
                print('Creating a new brain from scratch.')
        except IOError:
            print('Couldn\'t open {0} for loading'.format(
                self.pickle_filename))
        except pickle.PickleError, err:
            print('Error unpickling world: {0}'.format(err))
        return restored_brain


    def visualize(self):
        """
        Show the current state and some history of the brain.

        This is typically called from a world's visualize method.
        """
        print(' ')
        print('{0} is {1} time steps old'.format(self.name, self.timestep))

        self.affect.visualize(self.timestep, self.name, self.log_dir)
        #for level in self.levels:
        #    level.visualize()
